@@ -100,7 +100,10 @@ WheelHandler::WheelHandler(QObject *parent)
         setScrolling(false);
     });
 
+    m_xScrollAnimation.setEasingCurve(QEasingCurve::OutCubic);
     m_yScrollAnimation.setEasingCurve(QEasingCurve::OutCubic);
+    m_xInertiaScrollAnimation.setEasingCurve(QEasingCurve::OutQuad);
+    m_yInertiaScrollAnimation.setEasingCurve(QEasingCurve::OutQuad);
 
     connect(QGuiApplication::styleHints(), &QStyleHints::wheelScrollLinesChanged, this, [this](int scrollLines) {
         m_defaultPixelStepSize = 20 * scrollLines;
@@ -144,10 +147,22 @@ void WheelHandler::setTarget(QQuickItem *target)
 
     m_flickable = target;
     m_filterItem->setParentItem(target);
+    if (m_xScrollAnimation.targetObject()) {
+        m_xScrollAnimation.stop();
+    }
+    m_xScrollAnimation.setTargetObject(target);
     if (m_yScrollAnimation.targetObject()) {
         m_yScrollAnimation.stop();
     }
     m_yScrollAnimation.setTargetObject(target);
+    if (m_yInertiaScrollAnimation.targetObject()) {
+        m_yInertiaScrollAnimation.stop();
+    }
+    m_yInertiaScrollAnimation.setTargetObject(target);
+    if (m_xInertiaScrollAnimation.targetObject()) {
+        m_xInertiaScrollAnimation.stop();
+    }
+    m_xInertiaScrollAnimation.setTargetObject(target);
 
     if (target) {
         target->installEventFilter(this);
@@ -198,8 +213,15 @@ void WheelHandler::_k_rebindScrollBars()
         // We need to check if the parent inherits QQuickScrollView in case the
         // parent is another Flickable that already has a Kirigami WheelHandler.
         auto flickableParent = m_flickable->parentItem();
+        if (m_scrollView && m_scrollView != flickableParent) {
+            m_scrollView->removeEventFilter(this);
+        }
         if (flickableParent && flickableParent->inherits("QQuickScrollView")) {
-            const auto siblings = flickableParent->children();
+            if (m_scrollView != flickableParent) {
+                m_scrollView = flickableParent;
+                m_scrollView->installEventFilter(this);
+            }
+            const auto siblings = m_scrollView->children();
             for (const auto child : siblings) {
                 if (child->inherits("QQuickScrollBarAttached")) {
                     attachedToScrollView.attached = child;
@@ -388,25 +410,12 @@ void WheelHandler::classBegin()
     m_engine = qmlEngine(this);
     m_units = m_engine->singletonInstance<Kirigami::Platform::Units *>("org.kde.kirigami.platform", "Units");
     m_settings = m_engine->singletonInstance<Kirigami::Platform::Settings *>("org.kde.kirigami.platform", "Settings");
-    initSmoothScrollDuration();
-
-    connect(m_units, &Kirigami::Platform::Units::longDurationChanged, this, &WheelHandler::initSmoothScrollDuration);
-    connect(m_settings, &Kirigami::Platform::Settings::smoothScrollChanged, this, &WheelHandler::initSmoothScrollDuration);
 }
 
 void WheelHandler::componentComplete()
 {
 }
 
-void WheelHandler::initSmoothScrollDuration()
-{
-    if (m_settings->smoothScroll()) {
-        m_yScrollAnimation.setDuration(m_units->longDuration());
-    } else {
-        m_yScrollAnimation.setDuration(0);
-    }
-}
-
 void WheelHandler::setScrolling(bool scrolling)
 {
     if (m_wheelScrolling == scrolling) {
@@ -419,6 +428,84 @@ void WheelHandler::setScrolling(bool scrolling)
     m_filterItem->setEnabled(m_wheelScrolling);
 }
 
+void WheelHandler::startInertiaScrolling()
+{
+    const qreal width = m_flickable->width();
+    const qreal height = m_flickable->height();
+    const qreal contentWidth = m_flickable->property("contentWidth").toReal();
+    const qreal contentHeight = m_flickable->property("contentHeight").toReal();
+    const qreal topMargin = m_flickable->property("topMargin").toReal();
+    const qreal bottomMargin = m_flickable->property("bottomMargin").toReal();
+    const qreal leftMargin = m_flickable->property("leftMargin").toReal();
+    const qreal rightMargin = m_flickable->property("rightMargin").toReal();
+    const qreal originX = m_flickable->property("originX").toReal();
+    const qreal originY = m_flickable->property("originY").toReal();
+    const qreal contentX = m_flickable->property("contentX").toReal();
+    const qreal contentY = m_flickable->property("contentY").toReal();
+
+    QPointF minExtent = QPointF(leftMargin, topMargin) - QPointF(originX, originY);
+    QPointF maxExtent = QPointF(width, height) - (QPointF(contentWidth, contentHeight) + QPointF(rightMargin, bottomMargin) + QPointF(originX, originY));
+
+    QPointF totalDelta(0, 0);
+    for (const QPoint delta : m_wheelEvents) {
+        totalDelta += delta;
+    }
+    const uint64_t elapsed = std::max<uint64_t>(m_timestamps.first() - m_timestamps.last(), 1);
+
+    // The inertia is more natural if we multiply
+    // the actual scrolling speed by some factor,
+    // chosen manually here to be 2.5. Otherwise, the
+    // scrolling will appear to be too slow.
+    const qreal speedFactor = 2.5;
+
+    // We get the velocity in px/s by calculating
+    // displacement / elapsed time; we multiply by
+    // 1000 since the elapsed time is in ms.
+    QPointF vel = totalDelta * 1000 / elapsed * speedFactor;
+    QPointF startValue = QPointF(contentX, contentY);
+
+    // We decelerate at 4000px/s^2, chosen by manual test
+    // to be natural.
+    const qreal deceleration = 4000 * speedFactor;
+
+    // We use constant deceleration formulas to find:
+    // time = |velocity / deceleration|
+    // distance_traveled = time * velocity / 2
+    QPointF time = QPointF(qAbs(vel.x() / deceleration), qAbs(vel.y() / deceleration));
+    QPointF endValue = QPointF(startValue.x() + time.x() * vel.x() / 2, startValue.y() + time.y() * vel.y() / 2);
+
+    // We bound the end value so that we don't animate
+    // beyond the scrollable amount.
+    QPointF boundedEndValue =
+        QPointF(std::max(std::min(endValue.x(), -maxExtent.x()), -minExtent.x()), std::max(std::min(endValue.y(), -maxExtent.y()), -minExtent.y()));
+
+    // If we did bound the end value, we check how much
+    // (from 0 to 1) of the animation is actually played,
+    // and we adjust the time required for it accordingly.
+    QPointF progressFactor = QPointF((boundedEndValue.x() - startValue.x()) / (endValue.x() - startValue.x()),
+                                     (boundedEndValue.y() - startValue.y()) / (endValue.y() - startValue.y()));
+    // The formula here is:
+    // partial_time = complete_time * (1 - sqrt(1 - partial_progress_factor)),
+    // with partial_progress_factor being between 0 and 1.
+    // It can be obtained by inverting the OutQad easing formula,
+    // which is f(t) = t(2 - t).
+    // We also convert back from seconds to milliseconds.
+    QPointF realTime = QPointF(time.x() * (1 - std::sqrt(1 - progressFactor.x())), time.y() * (1 - std::sqrt(1 - progressFactor.y()))) * 1000;
+    m_wheelEvents.clear();
+    m_timestamps.clear();
+
+    m_xScrollAnimation.stop();
+    m_yScrollAnimation.stop();
+    m_xInertiaScrollAnimation.setStartValue(startValue.x());
+    m_xInertiaScrollAnimation.setEndValue(boundedEndValue.x());
+    m_xInertiaScrollAnimation.setDuration(realTime.x());
+    m_xInertiaScrollAnimation.start(QAbstractAnimation::KeepWhenStopped);
+    m_yInertiaScrollAnimation.setStartValue(startValue.y());
+    m_yInertiaScrollAnimation.setEndValue(boundedEndValue.y());
+    m_yInertiaScrollAnimation.setDuration(realTime.y());
+    m_yInertiaScrollAnimation.start(QAbstractAnimation::KeepWhenStopped);
+}
+
 bool WheelHandler::scrollFlickable(QPointF pixelDelta, QPointF angleDelta, Qt::KeyboardModifiers modifiers)
 {
     if (!m_flickable || (pixelDelta.isNull() && angleDelta.isNull())) {
@@ -440,7 +527,9 @@ bool WheelHandler::scrollFlickable(QPointF pixelDelta, QPointF angleDelta, Qt::K
     const qreal pageWidth = width - leftMargin - rightMargin;
     const qreal pageHeight = height - topMargin - bottomMargin;
     const auto window = m_flickable->window();
+    const auto screen = window ? window->screen() : nullptr;
     const qreal devicePixelRatio = window != nullptr ? window->devicePixelRatio() : qGuiApp->devicePixelRatio();
+    const qreal refreshRate = screen ? screen->refreshRate() : 0;
 
     // HACK: Only transpose deltas when not using xcb in order to not conflict with xcb's own delta transposing
     if (modifiers & m_defaultHorizontalScrollModifiers && qGuiApp->platformName() != QLatin1String("xcb")) {
@@ -450,78 +539,102 @@ bool WheelHandler::scrollFlickable(QPointF pixelDelta, QPointF angleDelta, Qt::K
 
     const qreal xTicks = angleDelta.x() / 120;
     const qreal yTicks = angleDelta.y() / 120;
-    qreal xChange;
-    qreal yChange;
     bool scrolled = false;
 
-    // Scroll X
-    if (contentWidth > pageWidth) {
+    auto getChange = [pageScrollModifiers = modifiers & m_pageScrollModifiers](qreal ticks, qreal pixelDelta, qreal stepSize, qreal pageSize) {
         // Use page size with pageScrollModifiers. Matches QScrollBar, which uses QAbstractSlider behavior.
-        if (modifiers & m_pageScrollModifiers) {
-            xChange = qBound(-pageWidth, xTicks * pageWidth, pageWidth);
-        } else if (pixelDelta.x() != 0) {
-            xChange = pixelDelta.x();
+        if (pageScrollModifiers) {
+            return qBound(-pageSize, ticks * pageSize, pageSize);
+        } else if (pixelDelta != 0) {
+            return pixelDelta;
         } else {
-            xChange = xTicks * m_horizontalStepSize;
+            return ticks * stepSize;
+        }
+    };
+
+    auto getPosition = [devicePixelRatio](qreal size,
+                                          qreal contentSize,
+                                          qreal contentPos,
+                                          qreal originPos,
+                                          qreal pageSize,
+                                          qreal leadingMargin,
+                                          qreal trailingMargin,
+                                          qreal change,
+                                          const QPropertyAnimation &animation) {
+        if (contentSize <= pageSize) {
+            return contentPos;
         }
 
         // contentX and contentY use reversed signs from what x and y would normally use, so flip the signs
 
-        qreal minXExtent = leftMargin - originX;
-        qreal maxXExtent = width - (contentWidth + rightMargin + originX);
-
-        qreal newContentX = qBound(-minXExtent, contentX - xChange, -maxXExtent);
-        // Flickable::pixelAligned rounds the position, so round to mimic that behavior.
-        // Rounding prevents fractional positioning from causing text to be
-        // clipped off on the top and bottom.
-        // Multiply by devicePixelRatio before rounding and divide by devicePixelRatio
-        // after to make position match pixels on the screen more closely.
-        newContentX = std::round(newContentX * devicePixelRatio) / devicePixelRatio;
-        if (contentX != newContentX) {
-            scrolled = true;
-            m_flickable->setProperty("contentX", newContentX);
-        }
-    }
-
-    // Scroll Y
-    if (contentHeight > pageHeight) {
-        if (modifiers & m_pageScrollModifiers) {
-            yChange = qBound(-pageHeight, yTicks * pageHeight, pageHeight);
-        } else if (pixelDelta.y() != 0) {
-            yChange = pixelDelta.y();
-        } else {
-            yChange = yTicks * m_verticalStepSize;
-        }
-
-        // contentX and contentY use reversed signs from what x and y would normally use, so flip the signs
-
-        qreal minYExtent = topMargin - originY;
-        qreal maxYExtent = height - (contentHeight + bottomMargin + originY);
-
-        qreal newContentY;
-        if (m_yScrollAnimation.state() == QPropertyAnimation::Running) {
-            m_yScrollAnimation.stop();
-            newContentY = std::clamp(m_yScrollAnimation.endValue().toReal() + -yChange, -minYExtent, -maxYExtent);
-        } else {
-            newContentY = std::clamp(contentY - yChange, -minYExtent, -maxYExtent);
-        }
+        qreal minExtent = leadingMargin - originPos;
+        qreal maxExtent = size - (contentSize + trailingMargin + originPos);
+        qreal newContentPos = (animation.state() == QPropertyAnimation::Running ? animation.endValue().toReal() : contentPos) - change;
+        // bound the values without asserts
+        newContentPos = std::max(-minExtent, std::min(newContentPos, -maxExtent));
 
         // Flickable::pixelAligned rounds the position, so round to mimic that behavior.
         // Rounding prevents fractional positioning from causing text to be
         // clipped off on the top and bottom.
         // Multiply by devicePixelRatio before rounding and divide by devicePixelRatio
         // after to make position match pixels on the screen more closely.
-        newContentY = std::round(newContentY * devicePixelRatio) / devicePixelRatio;
-        if (contentY != newContentY) {
-            scrolled = true;
-            if (m_wasTouched || !m_engine) {
-                m_flickable->setProperty("contentY", newContentY);
-            } else {
-                m_yScrollAnimation.setEndValue(newContentY);
-                m_yScrollAnimation.start(QAbstractAnimation::KeepWhenStopped);
-            }
+        return std::round(newContentPos * devicePixelRatio) / devicePixelRatio;
+    };
+
+    auto setPosition = [this, devicePixelRatio, refreshRate](qreal oldPos, qreal newPos, qreal stepSize, const char *property, QPropertyAnimation &animation) {
+        animation.stop();
+        if (oldPos == newPos) {
+            return false;
         }
-    }
+        if (!m_settings->smoothScroll() || !m_engine || refreshRate <= 0) {
+            animation.setDuration(0);
+            m_flickable->setProperty(property, newPos);
+            return true;
+        }
+
+        // Can't use wheelEvent->deviceType() to determine device type since
+        // on Wayland mouse is always regarded as touchpad:
+        // https://invent.kde.org/qt/qt/qtwayland/-/blob/e695a39519a7629c1549275a148cfb9ab99a07a9/src/client/qwaylandinputdevice.cpp#L445
+        // Mouse wheel can generate angle delta like 240, 360 and so on when
+        // scrolling very fast on some mice such as the Logitech M150.
+        // Mice with hi-res mouse wheels such as the Logitech MX Master 3 can
+        // generate angle deltas as small as 16.
+        // On X11, trackpads can also generate very fine angle deltas.
+
+        // Duration is based on the duration and movement for 120 angle delta.
+        // Shorten duration for smaller movements, limit duration for big movements.
+        // We don't want fine deltas to feel extra slow and fast scrolling should still feel fast.
+        // Minimum 3 frames for a 60hz display if delta > 2 physical pixels
+        // (start already rendered -> 1/3 rendered -> 2/3 rendered -> end rendered).
+        // Skip animation if <= 2 real frames for low refresh rate screens.
+        // Otherwise, we don't scale the duration based on refresh rate or
+        // device pixel ratio to avoid making the animation unexpectedly
+        // longer or shorter on different screens.
+
+        qreal absPixelDelta = std::abs(newPos - oldPos);
+        int duration = absPixelDelta * devicePixelRatio > 2 //
+            ? std::max(qCeil(1000.0 / 60.0 * 3), std::min(qRound(absPixelDelta * m_units->longDuration() / stepSize), m_units->longDuration()))
+            : 0;
+        animation.setDuration(duration <= qCeil(1000.0 / refreshRate * 2) ? 0 : duration);
+        if (animation.duration() > 0) {
+            animation.setStartValue(oldPos);
+            animation.setEndValue(newPos);
+            animation.start(QAbstractAnimation::KeepWhenStopped);
+        } else {
+            m_flickable->setProperty(property, newPos);
+        }
+        return true;
+    };
+
+    qreal xChange = getChange(xTicks, pixelDelta.x(), m_horizontalStepSize, pageWidth);
+    qreal newContentX = getPosition(width, contentWidth, contentX, originX, pageWidth, leftMargin, rightMargin, xChange, m_xScrollAnimation);
+
+    qreal yChange = getChange(yTicks, pixelDelta.y(), m_verticalStepSize, pageHeight);
+    qreal newContentY = getPosition(height, contentHeight, contentY, originY, pageHeight, topMargin, bottomMargin, yChange, m_yScrollAnimation);
+
+    // Don't use `||` because we need the position to be set for contentX and contentY.
+    scrolled |= setPosition(contentX, newContentX, m_horizontalStepSize, "contentX", m_xScrollAnimation);
+    scrolled |= setPosition(contentY, newContentY, m_verticalStepSize, "contentY", m_yScrollAnimation);
 
     return scrolled;
 }
@@ -577,6 +690,12 @@ bool WheelHandler::eventFilter(QObject *watched, QEvent *event)
         return false;
     }
 
+    // We only process keyboard events for QQuickScrollView.
+    const auto eventType = event->type();
+    if (item == m_scrollView && eventType != QEvent::KeyPress && eventType != QEvent::KeyRelease) {
+        return false;
+    }
+
     qreal contentWidth = 0;
     qreal contentHeight = 0;
     qreal pageWidth = 0;
@@ -589,7 +708,7 @@ bool WheelHandler::eventFilter(QObject *watched, QEvent *event)
     }
 
     // The code handling touch, mouse and hover events is mostly copied/adapted from QQuickScrollView::childMouseEventFilter()
-    switch (event->type()) {
+    switch (eventType) {
     case QEvent::Wheel: {
         // QQuickScrollBar::interactive handling Matches behavior in QQuickScrollView::eventFilter()
         if (m_filterMouseEvents) {
@@ -602,13 +721,6 @@ bool WheelHandler::eventFilter(QObject *watched, QEvent *event)
         }
         QWheelEvent *wheelEvent = static_cast<QWheelEvent *>(event);
 
-        // Can't use wheelEvent->deviceType() to determine device type since on Wayland mouse is always regarded as touchpad
-        // https://invent.kde.org/qt/qt/qtwayland/-/blob/e695a39519a7629c1549275a148cfb9ab99a07a9/src/client/qwaylandinputdevice.cpp#L445
-        // and we can only expect a touchpad never generates the same angle delta as a mouse
-
-        // mouse wheel can also generate angle delta like 240, 360 and so on when scrolling very fast
-        // only checking wheelEvent->angleDelta().y() because we only animate for contentY
-        m_wasTouched = (std::abs(wheelEvent->angleDelta().y()) != 0 && std::abs(wheelEvent->angleDelta().y()) % 120 != 0);
         // NOTE: On X11 with libinput, pixelDelta is identical to angleDelta when using a mouse that shouldn't use pixelDelta.
         // If faulty pixelDelta, reset pixelDelta to (0,0).
         if (wheelEvent->pixelDelta() == wheelEvent->angleDelta()) {
@@ -629,6 +741,17 @@ bool WheelHandler::eventFilter(QObject *watched, QEvent *event)
 
         Q_EMIT wheel(&m_kirigamiWheelEvent);
 
+        if (m_wheelEvents.count() > 6) {
+            m_wheelEvents.dequeue();
+            m_timestamps.dequeue();
+        }
+        if (m_wheelEvents.count() > 2 && wheelEvent->isEndEvent()) {
+            startInertiaScrolling();
+        } else {
+            m_wheelEvents.enqueue(wheelEvent->pixelDelta());
+            m_timestamps.enqueue(wheelEvent->timestamp());
+        }
+
         if (m_kirigamiWheelEvent.isAccepted()) {
             return true;
         }
